/*************************************************************************
 *
 * FILE  transfer.cpp
 * Copyright (c) 2006 Daniel Kahlin <daniel@kahlin.net>
 * Written by Daniel Kahlin <daniel@kahlin.net>
 * $Id: transfer.cpp,v 1.6 2006-10-21 14:51:20 tlr Exp $
 *
 * DESCRIPTION 
 *
 ******/

#include <algorithm>
#include <stdio.h>
#include "global.hpp"

#include "transfer.hpp"
// arch dependant
#include "timer.hpp"

static USleep sleeper;

Transfer::Transfer(Port& pt, int blocksize)
    : pt_(pt), blocksize_(blocksize)
{
}


void Transfer::read_mem(uint8_t mode, uint32_t base, uint32_t len, uint8_t* target)
{
    while (len) {
	uint16_t offs=static_cast<uint16_t>(base & 0x3fff);
	uint8_t  bank=static_cast<uint8_t>(base >> 14);
	uint16_t max_len=static_cast<uint16_t>(0x4000-offs);
	uint16_t block_len=std::min<int>(len, blocksize_);
	uint16_t tmp_len=std::min<int>(block_len, max_len);

	if (debug_g) {
	    printf("reading %s, bank=$%02X, offs=$%04X, len=$%04X (addr=$%06X)\n",(mode)?"ROM":"RAM",bank,offs,tmp_len,base);
	    fflush(stdout);
	}
	read_mem_int_(mode, bank, offs, tmp_len, target);
	base+=tmp_len;
	len-=tmp_len;
	target+=tmp_len;
    }
}

void Transfer::write_mem(uint8_t mode, uint32_t base, uint32_t len, uint8_t* source)
{
    while (len) {
	uint16_t offs=static_cast<uint16_t>(base & 0x3fff);
	uint8_t  bank=static_cast<uint8_t>(base >> 14);
	uint16_t max_len=static_cast<uint16_t>(0x4000-offs);
	uint16_t block_len=std::min<int>(len, blocksize_);
	uint16_t tmp_len=std::min<int>(block_len, max_len);

	if (debug_g) {
	    printf("writing %s, bank=$%02X, offs=$%04X, len=$%04X (addr=$%06X)\n",(mode)?"ROM":"RAM",bank,offs,tmp_len,base);
	    fflush(stdout);
	}
	write_mem_int_(mode, bank, offs, tmp_len, source);
	base+=tmp_len;
	len-=tmp_len;
	source+=tmp_len;
    }
}

void Transfer::execute_mem(uint16_t addr)
{
    // command execute_mem
    send_byte_(0x03);
    // addr
    send_byte_(addr & 0xff);
    send_byte_(addr >> 8);
}

void Transfer::read_mem_int_(uint8_t mode, uint8_t bank, uint16_t base, uint16_t len, uint8_t* target)
{
    // command read_mem
    send_byte_(0x01);

    // mode (0=ram, 1=rom)
    send_byte_(mode);

    // bank
    send_byte_(bank);

    // addr
    send_byte_(base & 0xff);
    send_byte_(base >> 8);

    // len 
    send_byte_(len & 0xff);
    send_byte_(len >> 8);


    uint8_t chk=0;
    for (int i=0; i < len; ++i) {
	uint8_t tmp=recv_byte_();
	chk+=tmp+1;
	target[i]=tmp;
    }
    uint8_t chkb=recv_byte_();

    if (chk != chkb)
	printf("ERROR: got block, chk=$%02X (expected $%02X)\n",chk,chkb);
}

void Transfer::write_mem_int_(uint8_t mode, uint8_t bank, uint16_t base, uint16_t len, uint8_t* source)
{
    // command write_mem
    send_byte_(0x02);

    // mode (0=ram, 1=rom (invalid))
    send_byte_(mode);

    // bank
    send_byte_(bank);

    // addr
    send_byte_(base & 0xff);
    send_byte_(base >> 8);

    // len 
    send_byte_(len & 0xff);
    send_byte_(len >> 8);

    uint8_t chk=0;
    for (int i=0; i < len; ++i) {
	uint8_t tmp=source[i];
	send_byte_(tmp);
	chk+=tmp+1;
    }
    send_byte_(chk);
    uint8_t chkb=recv_byte_();

    if (chk != chkb)
	printf("ERROR: sent block, chk=$%02X (receiver got $%02X)\n",chk,chkb);
}
 
void Transfer::send_byte_(uint8_t data)
{
    uint8_t tmp;
    // bit7-5 to d2/d1/d0, clk=0
    tmp=(data >> 5) & 0x07;
    tmp|=0x10;  // keep ACK high to be able to read it.
    pt_.write(tmp);
    while( pt_.read()&0x10 ); // wait for ACK low

    // bit4-2 to d2/d1/d0, clk=1
    tmp=(data >> 2) & 0x07;
    tmp|=0x08;
    tmp|=0x10;  // keep ACK high to be able to read it.
    pt_.write(tmp);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // bit1-0 to d1/d0  d2=0, clk=0
    tmp=(data >> 0) & 0x03;
    tmp|=0x10;  // keep ACK high to be able to read it.
    pt_.write(tmp);
    while( pt_.read()&0x10 );  // wait for ACK low

    // d2/d1/d0=1, clk=1
    tmp=0x0f;
    tmp|=0x10;  // keep ACK high to be able to read it.
    pt_.write(tmp);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // done!
}

uint8_t Transfer::recv_byte_()
{
    uint8_t data;
    // ack/d2/d1/d0=1, clk=0
    pt_.write(0x17);
    while( pt_.read()&0x10 ); // wait for ACK low
    sleeper.usleep(5);        // this is to allow for setup on the signals
    data=(pt_.read() & 0x07) << 5;
    
    // ack/d2/d1/d0=1, clk=1
    pt_.write(0x1f);
    while( !(pt_.read()&0x10) );  // wait for ACK high
    sleeper.usleep(5);        // this is to allow for setup on the signals
    data|=(pt_.read() & 0x07) << 2;

    // ack/d2/d1/d0=1, clk=0
    pt_.write(0x17);
    while( pt_.read()&0x10 );  // wait for ACK low
    sleeper.usleep(5);        // this is to allow for setup on the signals
    data|=(pt_.read() & 0x03) << 0;

    // ack/d2/d1/d0=1, clk=1
    pt_.write(0x1f);
    while( !(pt_.read()&0x10) );  // wait for ACK high

    // done!
    return data;
}

/* eof */
